"""
PyESD: More information abou the package that would be part of the metadata
"""

__version__ = "1.0.1"
__author__ = "Daniel Boateng"
__credits__ = " Earth System Dynamics Research Group of University of Tuebingen"
